/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */

package org.eclipse.jet.compiler;


import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.codegen.jet.JETException;
import org.eclipse.jet.internal.compiler.CommentElementDelegate;
import org.eclipse.jet.internal.compiler.DeclarationElementDelegate;
import org.eclipse.jet.internal.compiler.ErrorRedirectingCoreElementDelegate;
import org.eclipse.jet.internal.compiler.XMLElementDelegate;
import org.eclipse.jet.internal.l10n.JET2Messages;
import org.eclipse.jet.internal.parser.JETCoreElement;
import org.eclipse.jet.internal.parser.JETMark;
import org.eclipse.jet.internal.parser.JETParseEventListener2;
import org.eclipse.jet.internal.parser.JETParser;
import org.eclipse.jet.internal.parser.JETReader;
import org.eclipse.jet.internal.parser.ParseProblemSeverity;
import org.eclipse.jet.taglib.TagDefinition;
import org.eclipse.jet.taglib.TagLibraryManager;
import org.eclipse.jet.taglib.TagLibraryReference;


/**
 * JET Parser Listener used by the JET2 Compiler
 *
 */
class ASTCompilerParseListener implements JETParseEventListener2
{

  private static final String ID__ATTR = "id"; //$NON-NLS-1$

  private static final String PREFIX__ATTR = "prefix"; //$NON-NLS-1$

  private static final String TAGLIB__DIRECTIVE = "taglib"; //$NON-NLS-1$
  private static final String JET__DIRECTIVE = "jet"; //$NON-NLS-1$
  private final static class ElementStack
  {

    private final List stack = new ArrayList();

    /**
     * 
     */
    public ElementStack()
    {
      super();
    }

    /**
     * Push an XML element on the active element stack
     * @param element
     */
    public void push(XMLBodyElement element)
    {
      stack.add(element);
    }

    /**
     * Remove an XML element from the active element stack
     * @return the popped element
     * @throws IllegalStateException if the stack is empty
     */
    public XMLBodyElement pop()
    {
      if (stack.size() == 0)
      {
        throw new IllegalStateException("stack is empty"); //$NON-NLS-1$
      }
      return (XMLBodyElement)stack.remove(stack.size() - 1);
    }

    /**
     * Return the top element on the active element statkc
     * @return the top element
     * @throws IllegalStateException if the stack is empty
     */
    public XMLBodyElement peek()
    {
      if (stack.size() == 0)
      {
        throw new IllegalStateException("stack is empty"); //$NON-NLS-1$
      }
      return (XMLBodyElement)stack.get(stack.size() - 1);

    }

    /**
     * Test if the stack is empty
     * @return <code>true</code> if empty, <code>false</code> otherwise
     */
    public boolean isEmpty()
    {
      return stack.size() == 0;
    }

    /**
     * Find the index of the element with the specified name that is closest to the top of the stack.
     * @param name the element name to search for
     * @return the found element's index or -1 if not found
     */
    public int findElementIndex(String name)
    {
      for (int i = stack.size() - 1; i >= 0; i--)
      {
        XMLBodyElement element = get(i);
        if (element.getName().equalsIgnoreCase(name))
        {
          return i;
        }
      }

      return -1;
    }

    /**
     * Test if the passed index is that of the top-most element in the active element stack.
     * @param index the test index
     * @return <code>true</code> if the index represents the top-most element, <code>false</code> otherwise.
     * @throws IllegalArgumentException if the index is less than 0 or larger than the stack size.
     */
    public boolean isAtTop(int index)
    {
      if (index < 0 || index >= stack.size())
      {
        throw new IllegalArgumentException();
      }
      return index == stack.size() - 1;
    }

    /**
     * Return the element at the specified index
     * @param index
     * @return the element
     * @throws IllegalArgumentException if the index is less than 0 or larger than the stack size.
     * 
     */
    public XMLBodyElement get(int index)
    {
      if (index < 0 || index >= stack.size())
      {
        throw new IllegalArgumentException();
      }
      return (XMLBodyElement)stack.get(index);
    }
  }

  /**
   * Stack of elements waiting for end tags.
   */
  private final ElementStack elementStack = new ElementStack();

  private JET2CompilationUnit compilationUnit;

  private JET2AST ast;

  private final TagLibraryUsageManager tagLibManager;

  private final JETReader reader;

  public ASTCompilerParseListener(JET2CompilationUnit unit, JETReader reader)
  {
    this(unit, reader, Collections.EMPTY_MAP);
  }

  public ASTCompilerParseListener(JET2CompilationUnit unit, JETReader reader, Map predefinedLibraryMap)
  {
    compilationUnit = unit;
    ast = compilationUnit.getAst();
    this.reader = reader;
    tagLibManager = new TagLibraryUsageManager(predefinedLibraryMap);
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener#beginPageProcessing()
   */
  public void beginPageProcessing()
  {
    // nothing to do
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener#handleDirective(java.lang.String, org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark, java.util.Map)
   */
  public void handleDirective(String directive, JETMark start, JETMark stop, Map attributes)
  {
    JET2Directive directiveImpl = ast.createDirective(start.getLine(), start.getCol(), start.getCursor(), stop.getCursor() + 1, directive, attributes);

    // although a directive may appear nested, it really isn't. Add it to the compilation unit.
    // TODO Does this work for @start and @end? Probably NOT.
    compilationUnit.getInternalBodyElements().addBodyElement(directiveImpl);

    if (TAGLIB__DIRECTIVE.equalsIgnoreCase(directive))
    {
      handleTagLibDirective(start, stop, attributes);
    }
    else if(JET__DIRECTIVE.equalsIgnoreCase(directive))
    {
      handleJetDirective(start, stop, attributes);
    } else {
      recordProblem(ParseProblemSeverity.WARNING, Problem.UnsupportedDirective, 
        JET2Messages.ASTCompilerParseListener_UnsupportedDirective,
        new Object[] {directive}, start.getCursor(), stop.getCursor(), start.getLine(), start.getCol() );
    }
  }


  private static Set knownJETAttributes = 
    new LinkedHashSet(Arrays.asList(new String[] {
      "skeleton", //$NON-NLS-1$
      "package", //$NON-NLS-1$
      "imports", //$NON-NLS-1$
      "class", //$NON-NLS-1$
      "nlString", //$NON-NLS-1$
      "startTag", //$NON-NLS-1$
      "endTag", //$NON-NLS-1$
      "version", //$NON-NLS-1$
    }));
  private static Set deprecatedJETAttributes =
    new LinkedHashSet(Arrays.asList(new String[] {
      "skeleton", //$NON-NLS-1$
      "nlString", //$NON-NLS-1$
    }));

  private JETParser parser;
  private void handleJetDirective(JETMark start, JETMark stop, Map attributes)
  {
    for (Iterator i = attributes.keySet().iterator(); i.hasNext();)
    {
      String attrName = (String)i.next();
      if(!knownJETAttributes.contains(attrName)) 
      {
        recordProblem(ParseProblemSeverity.ERROR, Problem.UnknownAttributeInTag, 
          JET2Messages.JET2Compiler_UnknownAttribute, new Object[] {attrName}, 
          start.getCursor(), stop.getCursor(), start.getLine(), start.getCol());
      }
      if(deprecatedJETAttributes.contains(attrName)) 
      {
        recordProblem(ParseProblemSeverity.WARNING, Problem.DeprecatedAttribute, 
          JET2Messages.JET2Compiler_DeprecatedAttribute, new Object[] {attrName}, 
          start.getCursor(), stop.getCursor(), start.getLine(), start.getCol());
      }
    }
    
    String pkg = (String)attributes.get("package"); //$NON-NLS-1$
    String cls = (String)attributes.get("class"); //$NON-NLS-1$
    String importStr = (String)attributes.get("imports"); //$NON-NLS-1$
    String startTag = (String)attributes.get("startTag"); //$NON-NLS-1$
    String endTag = (String)attributes.get("endTag"); //$NON-NLS-1$

    if(pkg != null) 
    {
      compilationUnit.setOutputJavaPackage(pkg);
    }
    if(cls != null)
    {
      compilationUnit.setOutputJavaClassName(cls);
    }
    if(importStr != null)
    {
      String[] imports = importStr.split("\\s+"); //$NON-NLS-1$
      compilationUnit.addImports(Arrays.asList(imports));
    }

    if(startTag != null)
    {
      parser.setStartTag(startTag);
    }
    if(endTag != null)
    {
      parser.setEndTag(endTag);
    }
  }

  /**
   * @param start
   * @param stop
   * @param attributes
   */
  private void handleTagLibDirective(JETMark start, JETMark stop, Map attributes)
  {
    String prefix = ((String)attributes.get(PREFIX__ATTR)).trim().toLowerCase();
    String id = (String)attributes.get(ID__ATTR);

    if (id == null)
    {
      compilationUnit.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.MissingRequiredAttribute,
        JET2Messages.JET2Compiler_MissingDirectiveAttribute,
        new Object []{ TAGLIB__DIRECTIVE, ID__ATTR },
        start.getCursor(),
        stop.getCursor(),
        start.getLine(), start.getCol());
    }
    else if (prefix == null)
    {
      compilationUnit.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.MissingRequiredAttribute,
        JET2Messages.JET2Compiler_MissingDirectiveAttribute,
        new Object []{ TAGLIB__DIRECTIVE, PREFIX__ATTR },
        start.getCursor(),
        stop.getCursor(),
        start.getLine(), start.getCol());
    }
    else if (!tagLibManager.canDefinePrefix(prefix, id))
    {
      compilationUnit.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.DuplicateXMLNamespacePrefix,
        JET2Messages.JET2Compiler_PrefixAlreadyAssigned,
        new Object []{ prefix, tagLibManager.getLibraryIdFromPrefix(prefix) },
        start.getCursor(),
        stop.getCursor(),
        start.getLine(), start.getCol());
    }
    else if (TagLibraryManager.getInstance().getTagLibrary(id) == null)
    {
      compilationUnit.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.UnknownTagLibrary,
        JET2Messages.JET2Compiler_UnknownTagLibrary,
        new Object []{ id },
        start.getCursor(),
        stop.getCursor(),
        start.getLine(), start.getCol());
    }
    else
    {
      tagLibManager.add(prefix, id);
    }
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener#handleExpression(org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark, java.util.Map)
   */
  public void handleExpression(JETMark start, JETMark stop, Map attributes)
  {
    JET2Expression expression = ast.createExpression(
      start.getLine(),
      start.getCol(),
      start.getCursor() - 3,
      stop.getCursor() + 2,
      start.getCursor(),
      stop.getCursor(), reader.getChars(start, stop));

    if (elementStack.isEmpty())
    {
      compilationUnit.getInternalBodyElements().addBodyElement(expression);
    }
    else
    {
      XMLBodyElement topElement = elementStack.peek();

      topElement.getInternalBodyElements().addBodyElement(expression);
    }

  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener#handleCharData(char[])
   */
  public void handleCharData(char[] chars)
  {
    TextElement text = ast.createTextElement(chars);

    if (elementStack.isEmpty())
    {
      compilationUnit.getInternalBodyElements().addBodyElement(text);
    }
    else
    {
      XMLBodyElement topElement = elementStack.peek();

      topElement.getInternalBodyElements().addBodyElement(text);
    }

  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener#endPageProcessing()
   */
  public void endPageProcessing()
  {
    while (!elementStack.isEmpty())
    {
      XMLBodyElement element = elementStack.pop();
      compilationUnit.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.MissingXmlEndTag,
        JET2Messages.JET2Compiler_MissingEndTag,
        new Object []{ element.getName() },
        element.getStart(),
        element.getEnd(),
        element.getLine(), element.getColumn());
    }

  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener#handleScriptlet(org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark, java.util.Map)
   */
  public void handleScriptlet(JETMark start, JETMark stop, Map attributes)
  {
    JET2Scriptlet scriplet = ast.createScriplet(
      start.getLine(),
      start.getCol(),
      start.getCursor() - 3,
      stop.getCursor() + 2,
      start.getCursor(),
      stop.getCursor(), reader.getChars(start, stop));

    if (elementStack.isEmpty())
    {
      compilationUnit.getInternalBodyElements().addBodyElement(scriplet);
    }
    else
    {
      XMLBodyElement topElement = elementStack.peek();

      topElement.getInternalBodyElements().addBodyElement(scriplet);
    }

  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener2#handleComment(org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark)
   */
  public void handleComment(JETMark start, JETMark stop)
  {
    Comment comment = ast.createComment(
      start.getLine(),
      start.getCol(),
      start.getCursor() - 4,
      stop.getCursor() + 4,
      start.getCursor(),
      stop.getCursor(), reader.getChars(start, stop));
    if (elementStack.isEmpty())
    {
      compilationUnit.getInternalBodyElements().addBodyElement(comment);
    }
    else
    {
      XMLBodyElement topElement = elementStack.peek();

      topElement.getInternalBodyElements().addBodyElement(comment);
    }
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener2#handleDeclaration(org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark)
   */
  public void handleDeclaration(JETMark start, JETMark stop)
  {
    JET2Declaration decl = ast.createDeclaration(
      start.getLine(),
      start.getCol(),
      start.getCursor() - 3,
      stop.getCursor() + 2,
      start.getCursor(),
      stop.getCursor(), reader.getChars(start, stop));

    compilationUnit.getInternalBodyElements().addBodyElement(decl);
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener2#handleXMLEndTag(java.lang.String, org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark)
   */
  public void handleXMLEndTag(String tagName, JETMark start, JETMark stop)
  {
    int tagIndex = elementStack.findElementIndex(tagName);
    XMLBodyElementEnd endTag = ast.createXMLBodyElementEnd(start.getLine(), start.getCol(), 
      start.getCursor(), stop.getCursor(), tagName);
    if (tagIndex == -1)
    {
      compilationUnit.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.MissingXmlStartTag,
        JET2Messages.JET2Compiler_MissingStartTag,
        new Object []{ tagName },
        start.getCursor(),
        stop.getCursor(),
        start.getLine(), start.getCol());
    }
    else
    {
      while (!elementStack.isAtTop(tagIndex))
      {
        XMLBodyElement top = elementStack.pop();
        compilationUnit.createProblem(
          ParseProblemSeverity.ERROR,
          Problem.MissingXmlEndTag,
          JET2Messages.JET2Compiler_MissingEndTag,
          new Object []{ top.getName() },
          start.getCursor(),
          stop.getCursor(),
          start.getLine(), start.getCol());
      }
      XMLBodyElement topElement = elementStack.pop();
      endTag.setStartTag(topElement);
      topElement.setEndTag(endTag);
    }
    if (elementStack.isEmpty())
    {
      compilationUnit.getInternalBodyElements().addBodyElement(endTag);
    }
    else
    {
      XMLBodyElement topElement = (XMLBodyElement)elementStack.peek();

      topElement.getInternalBodyElements().addBodyElement(endTag);
    }

  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener2#handleXMLEmptyTag(java.lang.String, org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark, java.util.Map)
   */
  public void handleXMLEmptyTag(String tagName, JETMark start, JETMark stop, Map attributeMap)
  {
    TagDefinition td = tagLibManager.getTagDefinition(tagName);

    XMLEmptyElement decl = ast.createXMLEmptyElement(
      start.getLine(),
      start.getCol(),
      start.getCursor(),
      stop.getCursor(),
      tagName,
      attributeMap,
      td);

    if (elementStack.isEmpty())
    {
      compilationUnit.getInternalBodyElements().addBodyElement(decl);
    }
    else
    {
      XMLBodyElement topElement = (XMLBodyElement)elementStack.peek();

      topElement.getInternalBodyElements().addBodyElement(decl);
    }
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener2#handleXMLStartTag(java.lang.String, org.eclipse.jet.internal.parser.JETMark, org.eclipse.jet.internal.parser.JETMark, java.util.Map)
   */
  public void handleXMLStartTag(String tagName, JETMark start, JETMark stop, Map attributeMap)
  {

    TagDefinition td = tagLibManager.getTagDefinition(tagName);

    if(!td.isContentAllowed() && tagName.indexOf(':') == -1)
    {
      // bug 147714: for DPTK compatibility, allow empty tags to appear as <tagname ...> instead of <tagname .../>
      // We only allow this for tags that have no namespace (as DPTK did not support namespaces).
      recordProblem(ParseProblemSeverity.WARNING, Problem.TagInterpretedAsEmptyTag,
        JET2Messages.JET2Compiler_TagShouldBeEmptyFormat,
        new Object [] {"<" + tagName + "/>"},  //$NON-NLS-1$ //$NON-NLS-2$
        start.getCursor(), stop.getCursor(), start.getLine(), start.getCol());
      handleXMLEmptyTag(tagName, start, stop, attributeMap);
    }
    else
    {
      XMLBodyElement decl = ast.createXMLBodyElement(
        start.getLine(),
        start.getCol(),
        start.getCursor(),
        stop.getCursor(),
        tagName,
        attributeMap,
        td);
  
      if (elementStack.isEmpty())
      {
        compilationUnit.getInternalBodyElements().addBodyElement(decl);
      }
      else
      {
        XMLBodyElement topElement = elementStack.peek();
  
        topElement.getInternalBodyElements().addBodyElement(decl);
      }
  
      elementStack.push(decl);
    }
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener2#isKnownTag(java.lang.String)
   */
  public boolean isKnownTag(String tagName)
  {
    return tagLibManager.isKnownTag(tagName);
  }

  /**
   * Return the compilation unit created as a result of handling the JET2 parser events.
   * @return compilation unit
   */
  public JET2CompilationUnit getCompilationUnit()
  {
    return compilationUnit;
  }

  /**
   * @see org.eclipse.jet.internal.parser.JETParseEventListener2#recordProblem(org.eclipse.jet.compiler.Problem.ParseProblemSeverity, int, java.lang.String, java.lang.Object[], int, int, int, int)
   */
  public void recordProblem(ParseProblemSeverity severity, int problemId, String message, Object[] msgArgs, int start, int end, int line, int colOffset)
  {
    compilationUnit.createProblem(severity, problemId, message, msgArgs, start, end, line, colOffset);
  }

  public TagLibraryReference[] getTagLibraryReferences()
  {
    TagLibraryReference[] result = tagLibManager.getTagLibraryReferences();
    
    return result;
  }
  
  public void parse(JETReader reader)  {
    JETParser.Directive directive = new JETParser.Directive();
    directive.getDirectives().add("jet"); //$NON-NLS-1$
    directive.getDirectives().add("taglib"); //$NON-NLS-1$
    directive.getDirectives().add("include"); //$NON-NLS-1$
    directive.getDirectives().add("start"); //$NON-NLS-1$
    directive.getDirectives().add("end"); //$NON-NLS-1$
    JETCoreElement[] coreElements = new JETCoreElement []{
      new ErrorRedirectingCoreElementDelegate(directive),
      new ErrorRedirectingCoreElementDelegate(new JETParser.Expression()),
      new ErrorRedirectingCoreElementDelegate(new CommentElementDelegate()),
      new ErrorRedirectingCoreElementDelegate(new DeclarationElementDelegate()),
      new ErrorRedirectingCoreElementDelegate(new JETParser.Scriptlet()),
      new ErrorRedirectingCoreElementDelegate(new XMLElementDelegate()), };
    parser = new JETParser(reader, this, coreElements);
    this.beginPageProcessing();
    try
    {
      parser.parse();
    }
    catch (JETException e)
    {
      // create a minimal compilation unit with the exeception recorded as the error.
      recordProblem(ParseProblemSeverity.ERROR, Problem.JETException, e.getLocalizedMessage(), null, 0, 0, 1, 1);
    }
    this.endPageProcessing();

  }

  public boolean isKnownInvalidTagName(String tagName)
  {
    return tagLibManager.isKnownInvalidTagName(tagName);
  }
}
