/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */

package org.eclipse.jet.compiler;


import java.text.MessageFormat;

import org.eclipse.jet.internal.parser.ParseProblemSeverity;


/**
 * Represent a compilation problem on a JET2 tempalte.
 *
 */
public class Problem
{

  /**
   * Error Id for an XML end tag that has no corresponding start tag.
   * @see #getId()
   */
  public static final int MissingXmlStartTag = 1;

  /**
   * Error Id for an XML start tag that has no corresponding end tag.
   * @see #getId()
   */
  public static final int MissingXmlEndTag = 2;

  /**
   * Error Id for an XML tag or JET directive that is missing a required attribute
   * @see #getId()
   */
  public static final int MissingRequiredAttribute = 3;

  /**
   * Error Id for taglib directive that defines a prefix defined by a preceding taglib directive
   * @see #getId()
   */
  public static final int DuplicateXMLNamespacePrefix = 4;

  /**
   * Error Id for taglib directive that defines references an unknown tag library id
   * @see #getId()
   */
  public static final int UnknownTagLibrary = 5;

  /**
   * Error Id for an attribute that is not defined in the tag definition
   */
  public static final int UnknownAttributeInTag = 6;

  /**
   * Represent an unterminated XML Tag
   */
  public static final int UnterminatedXMLTag = 7;

  /**
   * Represent a duplicate attribute in an XML Tag;
   */
  public static final int DuplicateAttribute = 8;

  /**
   * An underlying JETException was thrown by the JET parser
   */
  public static final int JETException = 9;

  /**
   * Two templates specify that they compile to the same Java Class
   */
  public static final int MultipleTemplatesWithSameJavaClass = 10;

  /**
   * Use of an attribute that has been deprecated.
   */
  public static final int DeprecatedAttribute = 11;

  /**
   * Tag may not have a body - the tag must be of the form &lt;tagName/&gt;.
   */
  public static final int TagCannotHaveContent = 12;

  /**
   * Tag must have content - the tag must be of the form &lt;tagName&gt;xxx&lt;/tagName&gt;.
   */
  public static final int TagCannotBeEmpty = 13;

  /**
   * Use of the tag has been deprecated.
   */
  public static final int DeprecatedTag = 14;

  /**
   * Unsupported Directive.
   */
  public static final int UnsupportedDirective = 15;

  /**
   * A tag that has a known tag library prefix, but is not a recognized name. Usually
   * indicates a typographical error.
   */
  public static final int UnknownXMLTag = 16;

  /**
   * A tag this is declared as an 'emptyTag' occured as &lt;tag ...&gt;, and has been
   * interpreted as the equivalent empty tag &lt;tag .../&gt;.
   */
  public static final int TagInterpretedAsEmptyTag = 17;

  private final String originatingFileName;

  private final int id;

  private final String message;

  private final Object[] messageArgs;

  private final int start;

  private final int end;

  private final int lineNumber;

  private final ParseProblemSeverity severity;

  private final int colOffset;

  /**
   * Create an new instance
   * @param originatingFileName the file name from which the problem originates
   * @param severity the severity
   * @param id the problem id. See static constants declared on this class
   * @param message the error message in {@link MessageFormat} style
   * @param messageArgs the error message arguments
   * @param start the start offset of the problem (doc relative)
   * @param end the end offset of the problem (doc relative)
   * @param lineNumber the start line of the problem
   * @param colOffset TODO
   */
  public Problem(
    String originatingFileName,
    ParseProblemSeverity severity,
    int id,
    String message,
    Object[] messageArgs,
    int start,
    int end,
    int lineNumber, 
    int colOffset)
  {
    super();
    this.originatingFileName = originatingFileName;
    this.severity = severity;
    this.id = id;
    this.colOffset = colOffset;
    this.message = messageArgs != null && messageArgs.length > 0 ? MessageFormat.format(message, messageArgs) : message;
    this.messageArgs = messageArgs;
    this.start = start;
    this.end = end;
    this.lineNumber = lineNumber;
  }

  /**
   * @return Returns the end.
   */
  public int getEnd()
  {
    return end;
  }

  /**
   * @return Returns the id.
   */
  public int getId()
  {
    return id;
  }

  /**
   * @return Returns the lineNumber.
   */
  public int getLineNumber()
  {
    return lineNumber;
  }

  /**
   * @return Returns the message.
   */
  public String getMessage()
  {
    return message;
  }

  /**
   * @return Returns the messageArgs.
   */
  public Object[] getMessageArgs()
  {
    return messageArgs;
  }

  /**
   * @return Returns the originatingFileName.
   */
  public String getOriginatingFileName()
  {
    return originatingFileName;
  }

  /**
   * @return Returns the severity.
   */
  public ParseProblemSeverity getSeverity()
  {
    return severity;
  }

  /**
   * @return Returns the start.
   */
  public int getStart()
  {
    return start;
  }

  /**
   * @return Returns the colOffset.
   */
  public final int getColOffset()
  {
    return colOffset;
  }

}
