/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */

package org.eclipse.jet.compiler;


import java.util.Iterator;

import org.eclipse.jet.internal.l10n.JET2Messages;
import org.eclipse.jet.internal.parser.ParseProblemSeverity;
import org.eclipse.jet.taglib.TagAttributeDefinition;
import org.eclipse.jet.taglib.TagDefinition;


/**
 * Validate tags and attributes found on tag declarations. 
 * <p> 
 * Types of attribute validation:
 * <bl>
 * <li>that required attributes are in the tag declaration.</li>
 * <li>that all attributes correspond to a declared attribute.</li>
 * <li>that warnings are generated for deprecated attribute access.</li>
 * </bl>
 * <p>
 * Types of tag validation:
 * <bl>
 * <li>whether tag or tag library is deprecated</li>
 * <li>whether the tag is permitted to be used as an empty tag</li>
 * <li>whether the tag is permitted to be used as a content tag</li>
 * </bl>
 * </p>
 */
final class TagValidationVisitor extends DefaultJET2ASTVisitor implements JET2ASTVisitor
{

  private final JET2CompilationUnit cu;

  /**
   * @param cu 
   * 
   */
  public TagValidationVisitor(JET2CompilationUnit cu)
  {
    super();
    this.cu = cu;
  }

  public void visit(XMLBodyElement xmlBodyElement)
  {
    if (!xmlBodyElement.getTagDefinition().isContentAllowed())
    {
      cu.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.TagCannotHaveContent,
        JET2Messages.JET2Compiler_TagCannotHaveContent,
        new Object [0],
        xmlBodyElement.getStart(),
        xmlBodyElement.getEnd(),
        xmlBodyElement.getLine(), 
        xmlBodyElement.getColumn());
    }
    validateAttributes(xmlBodyElement);
  }

  private void validateAttributes(XMLElement xmlElement)
  {
    TagDefinition tagDefinition = xmlElement.getTagDefinition();

    if (tagDefinition.isDeprecated() || tagDefinition.getTagLibrary().isDeprecated())
    {
      cu.createProblem(
        ParseProblemSeverity.WARNING,
        Problem.DeprecatedTag,
        JET2Messages.JET2Compiler_DeprecatedTag,
        new Object [0],
        xmlElement.getStart(),
        xmlElement.getEnd(),
        xmlElement.getLine(), 
        xmlElement.getColumn());
    }

    // Validate that every supplied attribute is in the tag definition
    for (Iterator i = xmlElement.getAttributes().keySet().iterator(); i.hasNext();)
    {
      String attrName = (String)i.next();
      if (tagDefinition.getAttributeDefinition(attrName) == null)
      {
        cu.createProblem(
          ParseProblemSeverity.ERROR,
          Problem.UnknownAttributeInTag,
          JET2Messages.JET2Compiler_UnknownAttribute,
          new Object []{ attrName },
          xmlElement.getStart(),
          xmlElement.getEnd(),
          xmlElement.getLine(), 
          xmlElement.getColumn());
      }
    }

    // Validate that every required attribute is in the supplied attributes
    for (Iterator i = tagDefinition.getAttributeDefinitions().iterator(); i.hasNext();)
    {
      TagAttributeDefinition tad = (TagAttributeDefinition)i.next();
      final String attributeName = tad.getName();
      if (tad.isRequired() && !xmlElement.getAttributes().containsKey(attributeName))
      {
        cu.createProblem(
          ParseProblemSeverity.ERROR,
          Problem.MissingRequiredAttribute,
          JET2Messages.JET2Compiler_MissingRequiredAttribute,
          new Object []{ attributeName },
          xmlElement.getStart(),
          xmlElement.getEnd(),
          xmlElement.getLine(), 
          xmlElement.getColumn());
      }
      if (tad.isDeprecated() && xmlElement.getAttributes().containsKey(attributeName))
      {
        cu.createProblem(
          ParseProblemSeverity.WARNING,
          Problem.DeprecatedAttribute,
          JET2Messages.JET2Compiler_DeprecatedAttribute,
          new Object []{ attributeName },
          xmlElement.getStart(),
          xmlElement.getEnd(),
          xmlElement.getLine(), 
          xmlElement.getColumn());
      }
    }
  }

  public void visit(XMLEmptyElement xmlEmptyElement)
  {
    if (!xmlEmptyElement.getTagDefinition().isEmptyTagAllowed())
    {
      cu.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.TagCannotBeEmpty,
        JET2Messages.JET2Compiler_TagRequiresContent,
        new Object [0],
        xmlEmptyElement.getStart(),
        xmlEmptyElement.getEnd(),
        xmlEmptyElement.getLine(), 
        xmlEmptyElement.getColumn());
    }
    validateAttributes(xmlEmptyElement);
  }

}
