/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */
package org.eclipse.jet.compiler;

import org.eclipse.jet.internal.parser.LineInfo;


/**
 * Define a Text Element in the JET2 AST
 *
 */
public final class TextElement extends JET2ASTElement
{

  private static final String ESCAPED_JET_ELEMENT_START = "<\\%"; //$NON-NLS-1$

  private char[] text;

  private boolean trimLastLine = false;

  private boolean trimFirstLine = false;

  private final LineInfo[] lines;

  /**
   * Create an instance
   * @param jet2ast the AST
   * @param text the text
   */
  TextElement(JET2AST jet2ast, char[] text)
  {
    super(jet2ast, -1, -1, 0, 0);
    this.text = handleEscapes(text);
    this.lines = LineInfo.calculateLines(this.text);
  }

  /**
   * Remove any escape sequences in the raw text.
   * The only escaped sequences handled are:
   * <bl>
   * <li>&lt;\% to &lt;%</li>
   * </bl>
   * @param originalText
   * @return
   */
  private char[] handleEscapes(char[] originalText)
  {
    StringBuffer buffer = new StringBuffer(originalText.length);
    buffer.append(originalText);
    for (int i = buffer.indexOf(ESCAPED_JET_ELEMENT_START); i != -1; i = buffer.indexOf(ESCAPED_JET_ELEMENT_START, i))
    {
      buffer.replace(i, i + ESCAPED_JET_ELEMENT_START.length(), "<%"); //$NON-NLS-1$
    }
    return buffer.toString().toCharArray();
  }

  /**
   * @see org.eclipse.jet.compiler.JET2ASTElement#accept(org.eclipse.jet.compiler.JET2ASTVisitor)
   */
  public void accept(JET2ASTVisitor visitor)
  {
    visitor.visit(this);
  }

  /**
   * Return the text content
   * @return the text
   */
  public char[] getText()
  {
    char[] result = text;
    if(lines.length > 0)
    {
      int start = isTrimFirstLine() && lines[0].hasDelimiter() 
          ? lines[0].getEnd() + lines[0].getDelimiter().length() : 0;
      int end = isTrimLastLine() && !lines[lines.length -1].hasDelimiter() 
          ? lines[lines.length - 1].getStart() : text.length;
      result = new String(text, start, end - start).toCharArray();
    }
    return result;
  }
  
  public char[] getRawText()
  {
    return text;
  }

  /**
   * 
   * @param newText
   * @deprecated
   */
  void setText(char[] newText)
  {
    text = newText;
  }

  public boolean removeLineWhenOtherwiseEmpty()
  {
    return false;
  }

  public void setTrimLastLine(boolean trim)
  {
    this.trimLastLine = trim;
  }

  /**
   * @return Returns the trimLastLine.
   */
  public final boolean isTrimLastLine()
  {
    return trimLastLine;
  }

  public void setTrimFirstLine(boolean trim)
  {
    this.trimFirstLine = trim;
  }

  /**
   * @return Returns the trimFirstLine.
   */
  public final boolean isTrimFirstLine()
  {
    return trimFirstLine;
  }

  /**
   * @return Returns the lines.
   */
  public final LineInfo[] getLines()
  {
    return lines;
  }

}
